/*

  Program Hyperbolic_equation.ccp:
  Program to integrate a hyperbolic-parabolic partial differential equation
  through a second order Runge-Kutta scheme.

  d f / dt = - c * d f / dx - v * d^2 f / dx^2

  in an interval [a,b], with periodic boundary conditions,
  in the following cases:
  1) constant c;
  2) Burgers' equation;
  3) Upwind derivatives;

  The program reads the input data from file input.data.

 */

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
using namespace std;


// Output function, defined below the main program

void output( int, double, double *, double *, int );


/*

  Main program

*/

int main()
{

  /*
    Read initial data from file
  */

  ifstream finp;
  finp.open( "input.data" );

  // Read number of grid-points

  long int numPoints;
  finp >> numPoints;

  // Read the time step

  double dt;
  finp >> dt;

  // Read total time of the simulation

  double tEnd;
  finp >> tEnd;

  // Read time interval between the outputs

  double tPrint;
  finp >> tPrint;

  // Velocity parameter c (if not Burgers!)
  
  double c;
  finp >> c;

  // Dissipation parameter nu
  
  double nu;
  finp >> nu;

  // Generic parameter k (wavenumber of the initial condition)

  long int k;
  finp >> k;

  // Type of equation to integrate:
  // teq = 1 => constant velocity c = 1;
  // teq = 2 => Burgers' equation: c(x) = f(x);
  // teq = 3 => constant velocity, upwind derivatives.
  // teq = 4 => Burgers' equation, upwind derivatives.

  int teq;
  finp >> teq;

  finp.close();


  // Definition of the variables for the spatial grid a=-pi, b=pi

  double pi = acos( -1.0 );
  double a, b;
  a = -pi;
  b = +pi;


  // Variables for time loop

  long nEnd   = long( tEnd / dt );
  long nPrint = long( tPrint / dt );


  // Builds up the vectors for the spatial quantities

  double *x   = new double[numPoints];  // Grid-points
  double *f   = new double[numPoints];  // Solution at t_n and t_n+1
  double *fs  = new double[numPoints];  // Solution at t*
  double *d1f = new double[numPoints];  // First derivative of the solution
  double *d2f = new double[numPoints];  // Second derivative of the solution


  // Define the grid-points and the initial condition

  double tn = 0.0;
  int numPrint = 0;
  double dx = ( b - a ) / ( numPoints - 1 );
  double dx2 = dx * dx;
  double ddx = 2.0 * dx;
  for ( int j = 0; j < numPoints; j++ )
    {
      x[j] = a + j * dx;
      f[j] = - sin( k * x[j] );
    }


  // Output of the initial condition

  output( numPrint, tn, x, f, numPoints );

  
  // Time loop

  double cc;     // Variable to copy the value of the velocity for Burgers case
  for ( long int nt = 1; nt <= nEnd; nt++ )
    {

      // First step for the Runge-Kutta time scheme

      // Compute the first and second derivative according to the kind of equation

      // Internal points j=2,numPoints-1

      for ( int j = 1; j < numPoints - 1; j++ )
	{
	  switch ( teq )
	    {
	    case 1: // If not upwind derivatives, we use the centered scheme
	      d1f[j] = ( f[j+1] - f[j-1] ) / ddx;
	      break;
	    case 2: // If not upwind derivatives, we use the centered scheme
	      d1f[j] = ( f[j+1] - f[j-1] ) / ddx;
	      break;
	    case 3: // upwind derivatives, constant c
	      cc = c;
	      if ( cc > 0 ) d1f[j] = ( f[j] - f[j-1] ) / dx;
	      else d1f[j] = ( f[j+1] - f[j] ) / dx;
	      break;
	    case 4: // upwind derivatives, Burgers' equation
	      cc = f[j];
	      if ( cc > 0 ) d1f[j] = ( f[j] - f[j-1] ) / dx;
	      else d1f[j] = ( f[j+1] - f[j] ) / dx;
	      break;
	    }

	  // Second derivative, centered scheme
	  
	  d2f[j] = ( f[j+1] + f[j-1] - 2.0 * f[j] ) / dx2;

	}

      // Boundary points

      switch ( teq )
	{
	case 1: // If not upwind derivatives, we use the centered scheme
	  d1f[0] = ( f[1] - f[numPoints-2] ) / ddx;
	  d1f[numPoints-1] = ( f[1] - f[numPoints-2] ) / ddx;
	  break;
	case 2: // If not upwind derivatives, we use the centered scheme
	  d1f[0] = ( f[1] - f[numPoints-2] ) / ddx;
	  d1f[numPoints-1] = ( f[1] - f[numPoints-2] ) / ddx;
	  break;
	case 3: // upwind derivatives, constant c
	  cc = c;
	  if ( cc > 0 ) d1f[0] = ( f[0] - f[numPoints-2] ) / dx;
	  else d1f[0] = ( f[1] - f[0] ) / dx;
	  if ( cc > 0 ) d1f[numPoints-1] = ( f[numPoints-1] - f[numPoints-2] ) / dx;
	  else d1f[numPoints-1] = ( f[1] - f[numPoints-1] ) / dx;
	  break;
	case 4: // upwind derivatives, Burgers' equation
	  cc = f[0];
	  if ( cc > 0 ) d1f[0] = ( f[0] - f[numPoints-2] ) / dx;
	  else d1f[0] = ( f[1] - f[0] ) / dx;
	  cc = f[numPoints-1];
	  if ( cc > 0 ) d1f[numPoints-1] = ( f[numPoints-1] - f[numPoints-2] ) / dx;
	  else d1f[numPoints-1] = ( f[1] - f[numPoints-1] ) / dx;
	  break;
	}

      // Second derivative, centered scheme

      d2f[0] = ( f[1] + f[numPoints-2] - 2.0 * f[0] ) / dx2;
      d2f[numPoints-1] = ( f[1] + f[numPoints-2] - 2.0 * f[numPoints-1] ) / dx2;

      
      // Compute the intermediate value t*

      for ( int j = 0; j < numPoints; j++ )
	{
	  if ( ( teq == 2 ) || ( teq == 4 ) ) cc = f[j];
	  else cc = c;
	  fs[j] = f[j] + 0.5 * dt * ( - cc * d1f[j] + nu * d2f[j] );
	}
	

      // Second step for Runge-Kutta

      // Compute the second derivative according to b.c.

      // Internal points j=2,numPoints-1

      for ( int j = 1; j < numPoints - 1; j++ )
	{
	  switch ( teq )
	    {
	    case 1: // If not upwind derivatives, we use the centered scheme
	      d1f[j] = ( fs[j+1] - fs[j-1] ) / ddx;
	      break;
	    case 2: // If not upwind derivatives, we use the centered scheme
	      d1f[j] = ( fs[j+1] - fs[j-1] ) / ddx;
	      break;
	    case 3: // upwind derivatives, constant c
	      cc = c;
	      if ( cc > 0 ) d1f[j] = ( fs[j] - fs[j-1] ) / dx;
	      else d1f[j] = ( fs[j+1] - fs[j] ) / dx;
	      break;
	    case 4: // upwind derivatives, Burgers' equation
	      cc = fs[j];
	      if ( cc > 0 ) d1f[j] = ( fs[j] - fs[j-1] ) / dx;
	      else d1f[j] = ( fs[j+1] - fs[j] ) / dx;
	      break;
	    }

	  // Second derivative, centered scheme
	  
	  d2f[j] = ( fs[j+1] + fs[j-1] - 2.0 * fs[j] ) / dx2;

	}

      // Boundary points

      switch ( teq )
	{
	case 1: // If not upwind derivatives, we use the centered scheme
	  d1f[0] = ( fs[1] - fs[numPoints-2] ) / ddx;
	  d1f[numPoints-1] = ( fs[1] - fs[numPoints-2] ) / ddx;
	  break;
	case 2: // If not upwind derivatives, we use the centered scheme
	  d1f[0] = ( fs[1] - fs[numPoints-2] ) / ddx;
	  d1f[numPoints-1] = ( fs[1] - fs[numPoints-2] ) / ddx;
	  break;
	case 3: // upwind derivatives, constant c
	  cc = c;
	  if ( cc > 0 ) d1f[0] = ( fs[0] - fs[numPoints-2] ) / dx;
	  else d1f[0] = ( fs[1] - fs[0] ) / dx;
	  if ( cc > 0 ) d1f[numPoints-1] = ( fs[numPoints-1] - fs[numPoints-2] ) / dx;
	  else d1f[numPoints-1] = ( fs[1] - fs[numPoints-1] ) / dx;
	  break;
	case 4: // upwind derivatives, Burgers' equation
	  cc = fs[0];
	  if ( cc > 0 ) d1f[0] = ( fs[0] - fs[numPoints-2] ) / dx;
	  else d1f[0] = ( fs[1] - fs[0] ) / dx;
	  cc = fs[numPoints-1];
	  if ( cc > 0 ) d1f[numPoints-1] = ( fs[numPoints-1] - fs[numPoints-2] ) / dx;
	  else d1f[numPoints-1] = ( fs[1] - fs[numPoints-1] ) / dx;
	  break;
	}

      // Second derivative, centered scheme

      d2f[0] = ( fs[1] + fs[numPoints-2] - 2.0 * fs[0] ) / dx2;
      d2f[numPoints-1] = ( fs[1] + fs[numPoints-2] - 2.0 * fs[numPoints-1] ) / dx2;

      // Compute the final value at time step n+1

      for ( int j = 0; j < numPoints; j++ )
	{
	  if ( ( teq == 2 ) || ( teq == 4 ) ) cc = fs[j];
	  else cc = c;
	  f[j] = f[j] + dt * ( - cc * d1f[j] + nu * d2f[j] );
	}


      // Prints the outputs each Tprint time intervals

      tn = nt * dt;
      if ( ( nt % nPrint ) == 0 )
	{
	  numPrint++;
	  output( numPrint, tn, x, f, numPoints );
	}

    }

}




/*

  Function output

*/

void output( int np, double tt, double *x, double *f, int N )
{
  string fname = "out-" + to_string( np );
  string timestr = "# t = " + to_string( tt );
  ofstream fout;
  fout.open( fname );
  fout.precision( 15 );
  fout << timestr << endl;
  for ( int j = 0; j < N; j++ )
    fout << x[j] << "   " << f[j] << endl;
  fout.close();
}
