/*

  Program Parabolic_equation.ccp:
  Program to integrate a parabolic partial differential equation
  through a second order Runge-Kutta scheme.

  d f / dt = - v * d^2 f / dx^2

  in an interval [a,b] with boundary conditions:
  1) Dirichlet
  2) Neumann
  3) Periodic

  The program reads the input data from file input.data.

 */

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
using namespace std;


// Output function, defined below the main program

void output( int, double, double *, double *, int );


/*

  Main program

*/

int main()
{

  /*
    Read initial data from file
  */

  ifstream finp;
  finp.open( "input.data" );

  // Read number of grid-points

  long int numPoints;
  finp >> numPoints;

  // Read the time step

  double dt;
  finp >> dt;

  // Read total time of the simulation

  double tEnd;
  finp >> tEnd;

  // Read time interval between the outputs

  double tPrint;
  finp >> tPrint;

  // Dissipation parameter nu
  
  double nu;
  finp >> nu;

  // Generic parameter k (wavenumber of the initial condition)

  long int k;
  finp >> k;

  // Boundary conditions type:
  // bct = 1 => function vanishing on the boundaries (Dirichlet)
  // bct = 2 => first derivative vanishing on the boundaries (Neumann)
  // bct = 3 => periodic boundaries

  int bct;
  finp >> bct;

  finp.close();


  // Definition of the variables for the spatial grid a=-pi, b=pi

  double pi = acos( -1.0 );
  double a, b;
  a = -pi;
  b = +pi;


  // Variables for time loop

  long nEnd   = long( tEnd / dt );
  long nPrint = long( tPrint / dt );


  // Builds up the vectors for the spatial quantities

  double *x   = new double[numPoints];  // Grid-points
  double *f   = new double[numPoints];  // Solution at t_n and t_n+1
  double *fs  = new double[numPoints];  // Solution at t*
  double *d2f = new double[numPoints];  // Second derivative of the solution


  // Define the grid-points and the initial condition

  double tn = 0.0;
  int numPrint = 0;
  double dx = ( b - a ) / ( numPoints - 1 );
  double dx2 = dx * dx;
  for ( int j = 0; j < numPoints; j++ )
    {
      x[j] = a + j * dx;
      f[j] = 1.0 + cos( k * x[j] );    // Initial conditions for both Dirichlet (odd k) and Neumann (even k) b.c.
      //f[j] = 0.0;                      // Initial condition with a singularity in the middle of the domain
    }
  //f[(numPoints-1)/2] = 2.0;


  // Output of the initial condition

  output( numPrint, tn, x, f, numPoints );

  
  // Time loop

  for ( long int nt = 1; nt <= nEnd; nt++ )
    {

      // First step for the Runge-Kutta time scheme

      // Compute the second derivative according to b.c.

      // Internal points j=2,numPoints-1

      for ( int j = 1; j < numPoints - 1; j++ )
	{
	  d2f[j] = ( f[j+1] + f[j-1] - 2.0 * f[j] ) / dx2;
	}

      // Boundary points

      switch ( bct )
	{
	case 1: // f=0 on the boundary: any value is ok!
	  d2f[0] = 0.0;
	  d2f[numPoints-1] = 0.0;
	  break;
	case 2:
	  d2f[0] = 2.0 * ( f[1] - f[0] ) / dx2;
	  d2f[numPoints-1] = 2.0 * ( f[numPoints-2] - f[numPoints-1] ) / dx2;
	  break;
	case 3:
	  d2f[0] = ( f[1] + f[numPoints-2] - 2.0 * f[0] ) / dx2;
	  d2f[numPoints-1] = ( f[1] + f[numPoints-2] - 2.0 * f[numPoints-1] ) / dx2;
	  break;
	}

      
      // Compute the intermediate value t*

      for ( int j = 0; j < numPoints; j++ )
	fs[j] = f[j] + 0.5 * dt * nu * d2f[j];

      if ( bct == 1 )
	{
	  fs[0] = 0.0;
	  fs[numPoints-1] = 0.0;
	}

      
      // Second step for Runge-Kutta

      // Compute the second derivative according to b.c.

      // Internal points j=2,numPoints-1

      for ( int j = 1; j < numPoints - 1; j++ )
	{
	  d2f[j] = ( fs[j+1] + fs[j-1] - 2.0 * fs[j] ) / dx2;
	}

      // Boundary points

      switch ( bct )
	{
	case 1: // f=0 on the boundary: any value is ok!
	  d2f[0] = 0.0;
	  d2f[numPoints-1] = 0.0;
	  break;
	case 2:
	  d2f[0] = 2.0 * ( fs[1] - fs[0] ) / dx2;
	  d2f[numPoints-1] = 2.0 * ( fs[numPoints-2] - fs[numPoints-1] ) / dx2;
	  break;
	case 3:
	  d2f[0] = ( fs[1] + fs[numPoints-2] - 2.0 * fs[0] ) / dx2;
	  d2f[numPoints-1] = ( fs[1] + fs[numPoints-2] - 2.0 * fs[numPoints-1] ) / dx2;
	  break;
	}

      
      // Compute the final value at time step n+1

      for ( int j = 0; j < numPoints; j++ )
	f[j] = f[j] + dt * nu * d2f[j];


      // Forcing in the middle for the discontinuous solution
      //f[(numPoints-1)/2] = 2.0;


      // Prints the outputs each Tprint time intervals

      tn = nt * dt;
      if ( ( nt % nPrint ) == 0 )
	{
	  numPrint++;
	  output( numPrint, tn, x, f, numPoints );
	}

    }

}




/*

  Function output

*/

void output( int np, double tt, double *x, double *f, int N )
{
  string fname = "out-" + to_string( np );
  string timestr = "# t = " + to_string( tt );
  ofstream fout;
  fout.open( fname );
  fout.precision( 15 );
  fout << timestr << endl;
  for ( int j = 0; j < N; j++ )
    fout << x[j] << "   " << f[j] << endl;
  fout.close();
}
